/*
 * Handler routines.
 *
 * Copyright (C) 2015 Nicholas FitzRoy-Dale <wzdd.code@lardcave.net>
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

.text
.extern irq_handle
.extern current_stack_pointer
.extern current_task
.globl interrupt_handler
.set noreorder
.set noat

interrupt_handler:
	/* gp should have been set in the entrypoint. */

	jal save_state
	nop

	/* Record the stack pointer. */

	la $k0, current_stack_pointer
	sw $sp, 0($k0)

	/* Invoke the handler. */

	jal irq_handle
	nop

	/* Switch the stack pointer. */

	la $k0, current_stack_pointer
	lw $sp, 0($k0)

	/* Set the current task ASID. */

	la $k0, current_task
	lw $k1, 0($k0)
	mtc0 $k1, $10 /* CP0_ENTRYHI */
	nop

	j load_and_return
	nop

save_state:
	sw $at, -4($sp)
	sw $v0, -8($sp)
	sw $v1, -12($sp)
	sw $a0, -16($sp)
	sw $a1, -20($sp)
	sw $a2, -24($sp)
	sw $a3, -28($sp)
	sw $t0, -32($sp)
	sw $t1, -36($sp)
	sw $t2, -40($sp)
	sw $t3, -44($sp)
	sw $t4, -48($sp)
	sw $t5, -52($sp)
	sw $t6, -56($sp)
	sw $t7, -60($sp)
	sw $s0, -64($sp)
	sw $s1, -68($sp)
	sw $s2, -72($sp)
	sw $s3, -76($sp)
	sw $s4, -80($sp)
	sw $s5, -84($sp)
	sw $s6, -88($sp)
	sw $s7, -92($sp)
	sw $t8, -96($sp)
	/* sw $t9, -100($sp) */
	/* sw $gp, -104($sp) */
	sw $fp, -108($sp)
	/* sw $ra, -112($sp) */

	mfc0 $k0, $14 /* CP0_EPC */
	nop
	sw $k0, -116($sp)

	addi $sp, $sp, -120
	j $ra
	nop

load_and_return:

	addi $sp, $sp, 120

	lw $at, -4($sp)
	lw $v0, -8($sp)
	lw $v1, -12($sp)
	lw $a0, -16($sp)
	lw $a1, -20($sp)
	lw $a2, -24($sp)
	lw $a3, -28($sp)
	lw $t0, -32($sp)
	lw $t1, -36($sp)
	lw $t2, -40($sp)
	lw $t3, -44($sp)
	lw $t4, -48($sp)
	lw $t5, -52($sp)
	lw $t6, -56($sp)
	lw $t7, -60($sp)
	lw $s0, -64($sp)
	lw $s1, -68($sp)
	lw $s2, -72($sp)
	lw $s3, -76($sp)
	lw $s4, -80($sp)
	lw $s5, -84($sp)
	lw $s6, -88($sp)
	lw $s7, -92($sp)
	lw $t8, -96($sp)
	lw $t9, -100($sp)
	lw $gp, -104($sp)
	lw $fp, -108($sp)
	lw $ra, -112($sp)

	lw $k0, -116($sp)
	mtc0 $k0, $14 /* CP0_EPC */
	nop

	lw $k0, -120($sp)
	mtc0 $k0, $12 /* CP0_STATUS */
	nop

	eret
	nop

.set reorder
.set at
